--------------------------------------------------------------
-- 				USB Gecko v2.0 VHDL							--
--------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;

-- Define Ports
ENTITY usbgecko IS
    PORT(

-- USB Port   
		USBWR		: out std_logic := '0'; -- USB WR
		USBRD		: out std_logic := '1'; -- USB RD
		USBPWREN	: in std_logic; -- For USB Suspend logic only
		USBTXE		: in std_logic; -- USB TXE
		USBRXF		: in std_logic; -- USB RXF
		USBDATA 	: inout std_logic_vector(7 downto 0); -- USB Data Bus
-- Flash Port
		FlashAddress 	: out std_logic_vector(18 downto 0); -- Flash Address
		FlashData		: inout std_logic_vector(7 downto 0); -- Flash Data
		FlashWE			: out std_logic; -- Write Enable
		FlashCE			: out std_logic; -- Chip Enable
		FlashOE			: out std_logic; -- Output Enable
-- EXI Port
		EXICLK			: in std_logic; -- EXI Clock (Data seems to be valid on HIGH)
		EXICS			: in std_logic; -- EXI Chip Select
		EXIDI			: out std_logic; -- EXI DI
		EXID0			: in std_logic); -- EXI D0 (handles bi directional data)

END usbgecko;


ARCHITECTURE Behavioral OF usbgecko IS

-- USB Signals
signal usbstreambuffer	 : std_logic_vector(0 to 15); -- two byte version
signal datainbuffer  : std_logic_vector(7 downto 0); -- store usb recieve byte
signal usboktoread	 : std_logic;
signal usboktowrite	 : std_logic;
signal writetousb	 : std_logic;

-- Flash Signals
signal flashstreambuffer	: std_logic_vector(0 to 30);  -- two byte version
signal flashreadbuffer	 	: std_logic_vector(7 downto 0); 
signal flashwriteenable		: std_logic;
signal flashreadenable 		: std_logic;
signal flashaddressreg 		: std_logic_vector(18 downto 0);
signal flashdatareg 		: std_logic_vector(7 downto 0);
signal doflashwrite			: std_logic;
signal usbmode				: std_logic;

BEGIN   

	flashCE <= '0';		-- Chip Enable
--------------------------------------------------------
-- 	 Main Process for V2.0 No more state machine
--------------------------------------------------------
	
ReadEXIdata: Process(EXICLK,EXICS)	-- Main process (only one for v2.0) 
variable bitcount	: integer range 0 to 63;	-- use variable as just one process

begin
	
	if(EXICS = '1') then	-- Memory Card is deselected so reset values
	datainbuffer(7 downto 0) <= (others => '0');
	usboktoread <= '0';
	usboktowrite <= '0';
	writetousb <= '0';
	usbstreambuffer(0 to 15) <= (others => '0'); -- clear the buffer
	doflashwrite <= '0';
	flashwriteenable <= '0';
	flashreadenable <= '0';
	flashstreambuffer(0 to 30) <= (others => '0'); 
	flashreadbuffer(7 downto 0) <= (others => '0');
	FlashWE <= '1'; 	-- Write High
	FlashOE <= '1'; 	-- Output Enable High
	-- Reset the bit counter
	bitcount := 0;
	-- Reset the mode state (default flash / one is usb mode)
	usbmode <= '0';
	else
		-- Clock high so we do all our processing	
		if (EXICLK'event and EXICLK = '1') then -- Else CS is low lets go, do logic on clock high
		EXIDI <= '0';

		-- Tristate handler
		if(writetousb = '1') then
			USBDATA(7 downto 0) <= usbstreambuffer(4 to 11);
		else
			USBDATA(7 downto 0) <= (others => 'Z');
		end if;		
		
		-- Fill the buffers with stream data
		usbstreambuffer(bitcount) <= EXID0;
		flashstreambuffer(bitcount) <= EXID0;
		
		-- WE Handler
		if (doflashwrite = '1') then	
			FlashOE <= '1';
			flashWE <= '0';
			flashAddress(18 downto 0) <= flashaddressreg(18 downto 0); -- set address
			flashdata(7 downto 0) <= flashdatareg(7 downto 0); -- set data
		end if;
		
		if (doflashwrite = '0') then	
			FlashOE <= '0';
			flashWE <= '1';
			flashdata(7 downto 0) <= (others => 'Z'); -- tristate
		end if;

		------------------ Bit Count 3 ------------------------------
		if (bitcount = 3) then		
			if(usbstreambuffer(0 to 3) = "1010") then	-- 'A' command
				if(USBRXF = '0') then	-- If low then we can receive
					EXIDI <= '1';		-- so send a 1 	(0x04000000)
					usboktoread <= '1';	-- were ok to read
					USBRD <= '0';		-- USB low
				else
					EXIDI <= '0';		-- we cant read at this time
					usboktoread <= '0';	
					USBRD <= '1';		-- USB high (default)
				end if;
			end if;
		end if;

		------------------ Bit Count 4 ------------------------------
		if (bitcount = 4) then
			if(usbstreambuffer(0 to 3) = "1011") then	-- 'B' command	
				if(USBTXE = '0') then	-- If low then we can send
					EXIDI <= '1';		-- so send a 1 	(0x04000000)
					usboktowrite <= '1';	-- were ok to write
				else
					EXIDI <= '0';
					usboktowrite <= '0';	-- were ok to write
				end if;
			end if;
		
			if(usboktoread <= '1') then	-- if read signal is set we can read
				datainbuffer(7 downto 0) <= USBDATA(7 downto 0);
			end if;
			
			if(flashstreambuffer(0 to 3) = "1110") then --  E flash write mode
				flashwriteenable <= '1';
			else
				flashwriteenable <= '0';
			end if;
			
			if(flashstreambuffer(0 to 3) = "1111") then --  F flash read mode
				flashreadenable <= '1';
			else
				flashreadenable <= '0';
			end if;
		end if;
		------------------ Bit Count 5 ------------------------------
		if (bitcount = 5) then
			if(usbstreambuffer(0 to 3) = "1010") then	-- 'A' command
				USBRD <= '1';			-- back to high
			end if;
		end if;
		------------------ Bit Count 7 ------------------------------
		if (bitcount = 7) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(7);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
			
			if(flashwriteenable = '1' or flashreadenable = '1') then --  9 flash write mode		
				flashaddressreg(18 downto 16) <= flashstreambuffer(4 to 6);
			end if;	
		end if;
		------------------ Bit Count 8 ------------------------------
		if (bitcount = 8) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(6);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 9 ------------------------------
		if (bitcount = 9) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(5);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 10 ------------------------------
		if (bitcount = 10) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(4);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 11 ------------------------------
		if (bitcount = 11) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(3);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
		
			if(usboktowrite = '1') then	-- were still ready to write, lets send
				USBWR <= '1'; 			-- put high ready to write
			end if;		
		end if;
		
		------------------ Bit Count 12 ------------------------------
		if (bitcount = 12) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(2);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
			
			if(usboktowrite = '1') then	-- were still ready to write, lets send
				writetousb <= '1';
			end if;		
		end if;
		------------------ Bit Count 13 ------------------------------
		if (bitcount = 13) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(1);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
			
			if(usboktowrite = '1') then	-- were finished sending first byte
				USBWR <= '0'; 			-- put WR low
				writetousb <= '0';
			end if;		
		end if;
		------------------ Bit Count 14 ------------------------------
		if (bitcount = 14) then
			if(usboktoread = '1') then		-- ok to read flag still set?
				EXIDI <= datainbuffer(0);	-- start writing the buffer back
			else
				EXIDI <= '0';
			end if;
		
			if(flashwriteenable = '1' or flashreadenable = '1') then
				usbmode <= '0';
			else
				usbmode <= '1';
			end if;
		end if;
		------------------ Bit Count 15 ------------------------------
		if (bitcount = 15) then
			if(flashwriteenable = '1' or flashreadenable = '1') then
				flashaddressreg(15 downto 8) <= flashstreambuffer(7 to 14);
			end if;
		end if;
		------------------ Bit Count 23 ------------------------------
		if (bitcount = 23) then
			if(flashwriteenable = '1' or flashreadenable = '1') then
				flashaddressreg(7 downto 0) <= flashstreambuffer(15 to 22);
			end if;
		end if;
		------------------ Bit Count 30 ------------------------------
		if (bitcount = 30) then				
			if(flashreadenable = '1') then --  8 flash read mode
				flashAddress(18 downto 0) <= flashaddressreg(18 downto 0); -- set address
				flashreadbuffer(7 downto 0) <= flashdata(7 downto 0);
			end if;
		end if;
		------------------ Bit Count 31 ------------------------------
		if (bitcount = 31) then
			if(flashwriteenable = '1') then --  9 flash write mode
				flashdatareg(7 downto 0) <= flashstreambuffer(23 to 30);
			end if;
		
		if(flashreadenable = '1') then --  8 flash read mode
				-- data is read here also, it fixes bug
				flashAddress(18 downto 0) <= flashaddressreg(18 downto 0); -- set address
				flashreadbuffer(7 downto 0) <= flashdata(7 downto 0);
			end if;	
		end if;
		------------------ Bit Count 32 ------------------------------
		if (bitcount = 32) then
			if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(7);
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 33 ------------------------------
		if (bitcount = 33) then
			if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(6);
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 34 ------------------------------
		if (bitcount = 34) then
		if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(5);
			else
				EXIDI <= '0';
			end if;

			if(flashwriteenable = '1') then --  9 flash write mode
				doflashwrite <= '1';
			end if;
		end if;
		------------------ Bit Count 35 ------------------------------
		if (bitcount = 35) then
				if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(4);
			else
				EXIDI <= '0';
			end if;
		end if;

		------------------ Bit Count 36 ------------------------------
		if (bitcount = 36) then
			if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(3);
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 37 ------------------------------
		if (bitcount = 37) then
				if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(2);
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Bit Count 38 ------------------------------
		if (bitcount = 38) then
			if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(1);
			else
				EXIDI <= '0';
			end if;
		end if;
		
		if (bitcount = 39) then
		if (flashreadenable = '1') then
				EXIDI <= flashreadbuffer(0);
			else
				EXIDI <= '0';
			end if;
		end if;
		------------------ Mode Handler ------------------------------
		if (usbmode = '0') then
			bitcount := bitcount + 1;
		else
			bitcount := 0;
		end if;
		
		end if;	
	end if;
end process ReadEXIdata;

END Behavioral;
